#!/usr/bin/env python
'''module contains class for working with ELM327'''

import mod_globals

import serial                                    #sudo easy_install pyserial
import sys
import os
import io
import time
from datetime import datetime

snat = {"01":"760", "04":"762", "07":"771", "08":"778", "0D":"775", "0E":"76E", "0F":"770", 
		"13":"732", "1C":"76B", "1E":"768", "23":"773", "26":"765", "27":"76D", "29":"764", 
		"2A":"76F", "2C":"772", "3A":"7D2", "4D":"7BD", "50":"738", "51":"763", "57":"767", 
		"59":"734", "62":"7DD", "67":"793", "68":"77E", "6B":"7B5", "6E":"7E9", "77":"7DA", 
		"79":"7EA", "7A":"7E8", "7C":"77C", "86":"7A2", "87":"7A0", "93":"7BB", "95":"7EC", 
		"A5":"725", "A6":"726", "A8":"7B6", "C0":"7B9", "D1":"7EE", "F7":"736", "F8":"737", 
		"FA":"77B", "FD":"76F", "FE":"76C", "FF":"7D0"}
dnat = {"01":"740", "04":"742", "07":"751", "08":"758", "0D":"755", "0E":"74E", "0F":"750", 
		"13":"712", "1C":"74B", "1E":"748", "23":"753", "26":"745", "27":"74D", "29":"744", 
		"2A":"74F", "2C":"752", "3A":"7D6", "4D":"79D", "50":"718", "51":"743", "57":"747", 
		"59":"714", "62":"7DC", "67":"792", "68":"75A", "6B":"795", "6E":"7E1", "77":"7CA", 
		"79":"7E2", "7A":"7E0", "7C":"75C", "86":"782", "87":"780", "93":"79B", "95":"7E4", 
		"A5":"705", "A6":"706", "A8":"796", "C0":"799", "D1":"7E6", "F7":"716", "F8":"717", 
		"FA":"75B", "FD":"74F", "FE":"74C", "FF":"7D0"} 


class ELM:
  '''ELM327 class'''
  
  SerialPort = 0
  lf   = 0
  vf   = 0
  
  keepAlive   = 4			#send startSession to CAN after silence if startSession defined
  busLoad     = .200
  lastCMDtime = 0
  serialPortTimeout = 3
  
  buff = ""
  currentprotocol = ""
  currentaddress  = ""
  startSession    = ""
  
  rsp_cache = {}

    
  def __init__(self, port, speed, log, startSession='10C0'):
    
    if not mod_globals.opt_demo:
      self.SerialPort = serial.Serial(port, baudrate=speed, timeout=self.serialPortTimeout)
      
    if len(mod_globals.opt_log)>0:
      self.lf = open("./logs/elm_"+mod_globals.opt_log, "at")
      self.vf = open("./logs/ecu_"+mod_globals.opt_log, "at")
     
    self.lastCMDtime = 0
  
  def clear_cache(self):
    ''' Clear cache before screen update 
    '''    
    if not mod_globals.opt_demo:
      self.rsp_cache = {}
    
  def request(self, req, positive = '' ):
    ''' Check if request is saved in cache. 
    	If not then 
    	  - make real request
    	  - convert responce to one line
    	  - save in cache
    	returns response without consistency check  
    '''
    if req in self.rsp_cache.keys():
      return self.rsp_cache[req]
    
    rsp = self.cmd( req )
    
    res = ""
    for s in rsp.split('\n'):
      if ':' in s:
        res += s[2:].strip()+' '
      else:  #responce consists only from one frame
        if s.replace(' ','').startswith(positive):
          res += s.strip()+' '  
        
    rsp = res

    self.rsp_cache[req] = rsp

    if self.vf!=0:
      tmstr = datetime.now().strftime("%H:%M:%S.%f")
      self.vf.write(tmstr+";"+dnat[self.currentaddress]+";"+req+";"+rsp+"\n")
    
    return rsp

  def cmd(self, command):
    
    tb = time.time()   #time of beginning 
    tc = tb
    
    # If we are on CAN and there was more than keepAlive seconds of silence and 
    # start_session_can was executed then send startSession command again
    if ((tb-self.lastCMDtime)>self.keepAlive 
        and self.currentprotocol=="can" 
        and len(self.startSession)>0):
      
      if self.lf!=0:
        tmstr = datetime.now().strftime("%H:%M:%S.%f")
        self.lf.write("#["+tmstr+"]"+"KeepAlive\n")
      
      self.lastCMDtime = tc         # for not to get infinite loop
      if not mod_globals.opt_demo:
        self.SerialPort.write(str(self.startSession+"\r").encode("utf-8"))  #send keepalive  
      self.buff = ""
      while(True):         # wait responce
        
        if not mod_globals.opt_demo:
          byte = self.SerialPort.read()
        else:
          byte = '>'
        
        tc = time.time()
        if byte=='>': break
        if (tc-tb)>self.serialPortTimeout:
          self.buff += "TIMEOUT"
          break
    
    # Ensure time gap between commands
    if ((tb-self.lastCMDtime)<self.busLoad) and "AT" not in command.upper():
      time.sleep(self.busLoad-tb+self.lastCMDtime)
      
    if not mod_globals.opt_demo:
      #print str(command+"\r").encode("utf-8")
      self.SerialPort.write(str(command+"\r").encode("utf-8"))  #send command
    
    if self.lf!=0:
      tm = str(time.time())
      tmstr = datetime.now().strftime("%H:%M:%S.%f")
      self.lf.write(">["+tmstr+"]"+command+"\n")    

    self.buff = ""
    while(True):

      if not mod_globals.opt_demo:
        byte = self.SerialPort.read()
      else:
        byte = '>'
        
      if byte=='\r': byte='\n'
      self.buff += byte
      tc = time.time()
      if byte=='>':
        break
      if (tc-tb)>self.serialPortTimeout:
        self.buff += "TIMEOUT"
        break
        
    self.lastCMDtime = time.time()
              
    if self.lf!=0:
      tm = str(time.time())
      self.lf.write("<["+str(tc-tb)+"]"+self.buff+"\n")
    return self.buff

  def close_protocol(self):
    self.cmd("atpc")
  
  def start_session_can(self,start_session):
    self.startSession = start_session
    self.cmd(self.startSession)    

  def init_can(self):

    self.currentprotocol = "can"
    self.currentaddress  = "7e0"
    self.startSession = ""
    self.lastCMDtime = 0

    self.cmd("at z")
    self.cmd("at e1")
    self.cmd("at l1")
    self.cmd("at al")
    self.cmd("at st ff")
    self.cmd("at at 0")
    self.cmd("at sp 6")
    self.cmd("at at 1")
    self.lastCMDtime = 0
    
  def start_session_can(self,start_session):
    self.startSession = start_session
    self.cmd(self.startSession)    

  def init_iso(self):
    self.currentprotocol = "iso"
    self.currentaddress  = "7a"
    self.startSession = ""
    self.lastCMDtime = 0

    self.cmd("at z")
    self.cmd("at l1")
    self.cmd("at d1")
    self.cmd("at st ff")
    self.cmd("at at 0")
    self.cmd("at sp 5")
    #self.cmd("at at 1")
  
  def set_can_addr( self, addr ):
    
    if self.currentprotocol=="can" and self.currentaddress==addr:
      return

    self.currentprotocol = "can"
    self.currentaddress  = addr
    self.startSession = ""
    self.lastCMDtime = 0
      
    TXa = dnat[addr]
    RXa = snat[addr]
    
    self.cmd("at sh "+TXa)
    self.cmd("at cra "+RXa)
    self.cmd("at fc sh "+TXa)
    self.cmd("at fc sd 30 00 00") # status BS STmin
    self.cmd("at fc sm 1")
    self.cmd("at sp 6")
  
  def set_iso_addr( self, addr ):

    if self.currentprotocol=="iso" and self.currentaddress==addr:
      return

    self.currentprotocol = "iso"
    self.currentaddress  = addr
    self.startSession = ""
      
    self.cmd("at sh 82 "+addr+" f1")  #set address
    self.cmd("at sp 5")
  
